var ref = require("prosemirror-state");
var Selection = ref.Selection;
var NodeSelection = ref.NodeSelection;

var browser = require("./browser")

// Track the state of the current editor selection. Keeps the editor
// selection in sync with the DOM selection by polling for changes,
// as there is no DOM event for DOM selection changes.
var SelectionReader = function SelectionReader(view) {
  var this$1 = this;

  this.view = view

  // Track the state of the DOM selection.
  this.lastAnchorNode = this.lastHeadNode = this.lastAnchorOffset = this.lastHeadOffset = null
  this.lastSelection = view.state.selection
  this.poller = poller(this)

  view.content.addEventListener("focus", function () { return this$1.poller.receivedFocus(); })
  view.content.addEventListener("blur", function () { return this$1.poller.lostFocus(); })
};

SelectionReader.prototype.poll = function poll (origin) { this.poller.poll(origin) };

// : () → bool
// Whether the DOM selection has changed from the last known state.
SelectionReader.prototype.domChanged = function domChanged () {
  var sel = this.view.root.getSelection()
  return sel.anchorNode != this.lastAnchorNode || sel.anchorOffset != this.lastAnchorOffset ||
    sel.focusNode != this.lastHeadNode || sel.focusOffset != this.lastHeadOffset
};

// Store the current state of the DOM selection.
SelectionReader.prototype.storeDOMState = function storeDOMState (selection) {
  var sel = this.view.root.getSelection()
  this.lastAnchorNode = sel.anchorNode; this.lastAnchorOffset = sel.anchorOffset
  this.lastHeadNode = sel.focusNode; this.lastHeadOffset = sel.focusOffset
  this.lastSelection = selection
};

// : (?string) → bool
// When the DOM selection changes in a notable manner, modify the
// current selection state to match.
SelectionReader.prototype.readFromDOM = function readFromDOM (origin) {
  if (!this.view.hasFocus() || this.view.inDOMChange || !this.domChanged()) { return }

  var domSel = this.view.root.getSelection(), doc = this.view.state.doc
  var domNode = domSel.focusNode, head = this.view.docView.posFromDOM(domNode, domSel.focusOffset)
  var $head = doc.resolve(head), $anchor, selection
  if (domSel.isCollapsed) {
    $anchor = $head
    var nearestDesc = this.view.docView.nearestDesc(domNode)
    while (nearestDesc && !nearestDesc.node) { nearestDesc = nearestDesc.parent }
    if (nearestDesc && nearestDesc.node.isLeaf && NodeSelection.isSelectable(nearestDesc.node))
      { selection = new NodeSelection($head) }
  } else {
    $anchor = doc.resolve(this.view.docView.posFromDOM(domSel.anchorNode, domSel.anchorOffset))
  }

  if (!selection) {
    var bias = this.view.state.selection.head != null && this.view.state.selection.head < $head.pos ? 1 : -1
    selection = Selection.between($anchor, $head, bias)
  }
  if ($head.pos == selection.head && $anchor.pos == selection.anchor)
    { this.storeDOMState(selection) }
  this.view.props.onAction(selection.action(origin && {origin: origin}))
};
exports.SelectionReader = SelectionReader

function poller(reader) {
  // There's two polling models. On browsers that support the
  // selectionchange event (everything except Firefox, basically), we
  // register a listener for that whenever the editor is focused.
  if ("onselectionchange" in document) { return new (function () {
      function anonymous() {
    var this$1 = this;

      this.listening = false
      this.curOrigin = null
      this.originTime = 0

      this.readFunc = function () { return reader.readFromDOM(this$1.originTime > Date.now() - 50 ? this$1.curOrigin : null); }
    }

    anonymous.prototype.poll = function poll (origin) {
      this.curOrigin = origin
      this.originTime = Date.now()
    };

    anonymous.prototype.receivedFocus = function receivedFocus () {
      if (!this.listening) {
        document.addEventListener("selectionchange", this.readFunc)
        this.listening = true
      }
    };

    anonymous.prototype.lostFocus = function lostFocus () {
      if (this.listening) {
        document.removeEventListener("selectionchange", this.readFunc)
        this.listening = false
      }
    };

      return anonymous;
    }()) }
  // On Firefox, we use timeout-based polling.
  return new (function () {
    function anonymous$1() {
      // The timeout ID for the poller when active.
      this.polling = null
      this.reader = reader
      this.pollFunc = this.doPoll.bind(this, null)
    }

    anonymous$1.prototype.doPoll = function doPoll (origin) {
      if (this.reader.view.hasFocus()) {
        this.reader.readFromDOM(origin)
        this.polling = setTimeout(this.pollFunc, 100)
      } else {
        this.polling = null
      }
    };

    anonymous$1.prototype.poll = function poll (origin) {
      clearTimeout(this.polling)
      this.polling = setTimeout(origin ? this.doPoll.bind(this, origin) : this.pollFunc, 0)
    };

    anonymous$1.prototype.receivedFocus = function receivedFocus () {
      if (this.polling == null) { this.poll() }
    };

    anonymous$1.prototype.lostFocus = function lostFocus () {
      clearTimeout(this.polling)
      this.polling = null
    };

    return anonymous$1;
  }())
}

function selectionToDOM(view, sel, takeFocus) {
  syncNodeSelection(view, sel)

  if (!view.hasFocus()) {
    if (!takeFocus) { return }
    // See https://bugzilla.mozilla.org/show_bug.cgi?id=921444
    else if (browser.gecko) { view.content.focus() }
  }

  var reader = view.selectionReader
  if (sel.eq(reader.lastSelection) && !reader.domChanged()) { return }
  var anchor = sel.anchor, head = sel.head
  if (anchor == null) { anchor = sel.from; head = sel.to }
  view.docView.setSelection(anchor, head, view.root)
  reader.storeDOMState(sel)
}
exports.selectionToDOM = selectionToDOM

function syncNodeSelection(view, sel) {
  if (sel instanceof NodeSelection) {
    var desc = view.docView.descAt(sel.from)
    if (desc != view.lastSelectedViewDesc) {
      clearNodeSelection(view)
      if (desc) {
        view.content.classList.add("ProseMirror-nodeselection")
        desc.selectNode()
      }
      view.lastSelectedViewDesc = desc
    }
  } else {
    clearNodeSelection(view)
  }
}

// Clear all DOM statefulness of the last node selection.
function clearNodeSelection(view) {
  if (view.lastSelectedViewDesc) {
    view.content.classList.remove("ProseMirror-nodeselection")
    view.lastSelectedViewDesc.deselectNode()
    view.lastSelectedViewDesc = null
  }
}
